/* -*- indent-tabs-mode: nil; sh-basic-offset: 4; -*-
 * vim: shiftwidth=4 softtabstop=4 expandtab
 *    Filename: librt.h
 *      Author: Darren Hart <dvhltc@us.ibm.com>
 * Description: Some basic statistical analysis convenience tools.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 * Copyright (C) IBM Corporation, 2006
 *
 * 2006-Oct-17:	Initial version by Darren Hart
 *
 * TODO: the save routine for gnuplot plotting should be more modular...
 */

#ifndef LIBSTATS_H
#define LIBSTATS_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <math.h>

#define MIN(A,B) ((A)<(B)?(A):(B))
#define MAX(A,B) ((A)>(B)?(A):(B))

typedef struct stats_record {
	long x;
	long y;
} stats_record_t;

typedef struct stats_container {
	long size;
	stats_record_t *records;
} stats_container_t;

/* function prototypes */

/* stats_container_create - allocate memory for a new container
 * size: number of records to allocate
 * data: stats_container_t destination pointer
 */
int stats_container_init(stats_container_t *data, long size);

/* stats_container_free - free the records array
 * data: stats_container_t to free records
 */
int stats_container_free(stats_container_t *data);

/* stats_stddev - return the standard deviation of the y values in data
 * data: stats_container_t data with y values for use in the calculation
 */
float stats_stddev(stats_container_t *data);

/* stats_avg - return the average (mean) of the y values in data
 * data: stats_container_t data with y values for use in the calculation
 */
float stats_avg(stats_container_t *data);

/* stats_hist - calculate a histogram with hist->size divisions from data
 * hist: the destination of the histogram data
 * data: the source from which to calculate the histogram
 */
int stats_hist(stats_container_t *hist, stats_container_t *data);

/* stats_hist_print - print out a histogram in human readable format
 * hist: the stats_container containing the histogram data
 */
void stats_hist_print(stats_container_t *hist);

/* stats_container_save - save the x,y data to a file and create a gnuplot 
 * runnable script.
 * filename: the filename to save the data as without an extension. A .dat
 * and a .plt file will be created.
 * title: the title of the graph
 * labelx: the x-axis label
 * labely: the y-axis label
 * mode: "points" "lines" "steps" etc, see gnuplot help for plotting types
 */
int stats_container_save(char *filename, char *title, char *labelx, char *labely, stats_container_t *data, char *mode);

/* function implementations */
int stats_container_init(stats_container_t *data, long size)
{
	data->size = size;
	data->records = (stats_record_t *)malloc(size*sizeof(stats_record_t));
	memset(data->records, 0, size*sizeof(stats_record_t));
}

int stats_container_free(stats_container_t *data)
{
	free(data->records);
}

float stats_stddev(stats_container_t *data)
{	
	long i;
	float sd, avg, sum, delta;
	long n;

	sd = 0.0;
	n = data->size;
	sum = 0.0;
	
	/* calculate the mean */
	for (i = 0; i < n; i++) {
		sum += data->records[i].y;
	}
	avg = sum / (float)n;

	/* calculate the standard deviation */
	sum = 0.0;
	for (i = 0; i < n; i++) {
		delta = (data->records[i].y - avg);
		sum += delta*delta;
	}
	sum /= n;

	sd = sqrt(sum);

	return sd;
}

float stats_avg(stats_container_t *data)
{	
	long i;
	float avg, sum;
	long n;

	n = data->size;
	sum = 0.0;
	
	/* calculate the mean */
	for (i = 0; i < n; i++) {
		sum += data->records[i].y;
	}
	avg = sum / (float)n;

	return avg;
}

int stats_hist(stats_container_t *hist, stats_container_t *data)
{
	int i;
	int ret;
	long min, max, width;
	long x, y, b;
	
	ret = 0;

	if (hist->size <= 0 || data <= 0) {
		return -1;
	}

	/* calculate the range of dataset */
	min = max = data->records[0].y;
	for (i = 0; i < data->size; i++) {
		y = data->records[i].y;
		if (y > max) max = y;
		if (y < min) min = y;
	}
	//printf("min = %d\n", min);
	//printf("max = %d\n", max);

	/* define the bucket ranges */
	width = MAX((max-min)/hist->size, 1);
	//printf("width = %d\n", width);
	hist->records[0].x = min;
	//printf("hist[0].x = %d\n", min);
	for (i = 1; i < (hist->size); i++) {
		hist->records[i].x = min + i*width;
		//printf("hist[%d].x = %d\n", i, hist->records[i].x);
	}

	/* fill in the counts */
	for (i = 0; i < data->size; i++) {
		y = data->records[i].y;
		b = MIN((y-min)/width, hist->size-1);
		//printf("%d will go in bucket %d\n", y, b);
		hist->records[b].y++;
	}

	return 0;
}

void stats_hist_print(stats_container_t *hist)
{
	long i, x, x1;
	for (i = 0; i < hist->size; i++) {
		x = hist->records[i].x;
		if (i < hist->size-1)
			printf("[%d,%d) = %d\n", x, 
				hist->records[i+1].x, hist->records[i].y);
		else
			printf("[%d,-) = %d\n", x, hist->records[i].y);
	}
}

int stats_container_save(char *filename, char *title, char *xlabel, char *ylabel, stats_container_t *data, char *mode)
{
    int ret;
    int i;
    int minx, maxx, miny, maxy;
    FILE *dat_fd;
    FILE *plt_fd;
    char *datfile;
    char *pltfile;
    stats_record_t *rec;

    /* generate the filenames */
    datfile = (char *)malloc(strlen(filename) + 4);
    strcpy(datfile, filename);
    strcpy(datfile + strlen(datfile), ".dat");
    pltfile = (char *)malloc(strlen(filename) + 4);
    strcpy(pltfile, filename);
    strcpy(pltfile + strlen(pltfile), ".plt");

    /* generate the data file */
    if (!(dat_fd = fopen(datfile, "w"))) {
        perror("Failed to open dat file");
        ret = -1;
    } else {
        minx = maxx = data->records[0].x;
        miny = maxy = data->records[0].y;
        for (i = 0; i < data->size; i++) {
            rec = &data->records[i];
            minx = MIN(minx, rec->x);            
            maxx = MAX(maxx, rec->x);            
            miny = MIN(miny, rec->y);            
            maxy = MAX(maxy, rec->y);            
            fprintf(dat_fd, "%d %d\n", rec->x, rec->y);
        }
        fclose(dat_fd);
    } 

    /* generate the plt file */
    if (!(plt_fd = fopen(pltfile, "w"))) {
        perror("Failed to open plt file");
        ret = -1;
    } else {
        fprintf(plt_fd, "set terminal png\n");
        fprintf(plt_fd, "set output \"%s.png\"\n", pltfile);
        fprintf(plt_fd, "set title \"%s\"\n", title);
        fprintf(plt_fd, "set xlabel \"%s\"\n", xlabel);
        fprintf(plt_fd, "set ylabel \"%s\"\n", ylabel);

        // exact range mode
        //fprintf(plt_fd, "plot [%d:%d] [%d:%d] \"%s\" with %s\n", 
        //        minx, maxx, miny, maxy, datfile, mode);

        // expanded range mode - slightly more intuitive I think...
        fprintf(plt_fd, "plot [0:%d] [0:%d] \"%s\" with %s\n", 
                maxx+1, maxy+1, datfile, mode);
        fclose(plt_fd);
    } 

    return ret;
}
#endif /* LIBSTAT_H */
